"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Schema = void 0;
// Copyright (c) Microsoft Corporation. All rights reserved.
// Licensed under the MIT License.
var SchemaClass_1 = require("./SchemaClass");
var SchemaEnum_1 = require("./SchemaEnum");
var fs = require("fs");
var path = require("path");
var Schema = /** @class */ (function () {
    function Schema(types) {
        var _this = this;
        this._typeDictionary = new Map();
        types.forEach(function (type) {
            _this._typeDictionary.set(type.type, _this.parse(type));
        });
        // Resolve types
        this._typeDictionary.forEach(function (type) {
            type.resolve(_this._typeDictionary);
        });
        this._typeDictionary.forEach(function (type) {
            if (type instanceof SchemaClass_1.SchemaClass) {
                type.resolveImplementations(_this._typeDictionary);
            }
        });
    }
    Schema.prototype.parse = function (sourceObj) {
        if (sourceObj.classType === "Enum") {
            return new SchemaEnum_1.SchemaEnum(sourceObj);
        }
        else if (sourceObj.classType === "Class" || sourceObj.classType === undefined) {
            return new SchemaClass_1.SchemaClass(sourceObj);
        }
        else {
            throw new Error("Unknown classType " + sourceObj.classType);
        }
    };
    Object.defineProperty(Schema.prototype, "typeDictionary", {
        get: function () {
            return this._typeDictionary;
        },
        enumerable: false,
        configurable: true
    });
    Schema.prototype.getType = function (typeName) {
        var answer = this._typeDictionary.get(typeName);
        if (answer === undefined) {
            var knownTypes = "";
            this._typeDictionary.forEach(function (value, key) {
                if (knownTypes.length > 0) {
                    knownTypes += ", ";
                }
                knownTypes += key;
            });
            throw new Error("Type " + typeName + " could not be found. Known types: " + knownTypes);
        }
        return answer;
    };
    Schema.fromFolder = function (folder) {
        var files = Schema.getAllFiles(folder);
        var types = [];
        files.forEach(function (filePath) {
            if (filePath.endsWith(".json")) {
                var fileTxt = fs.readFileSync(filePath, "utf8");
                var type = JSON.parse(fileTxt);
                // Infer type name from file name if not specified
                if (!type.type) {
                    var stat = fs.statSync(filePath);
                    if (stat) {
                        type.type = path.basename(filePath, ".json");
                    }
                }
                types.push(type);
                types[type.type] = type;
            }
        });
        return new Schema(types);
    };
    Schema.getAllFiles = function (dir) {
        // https://stackoverflow.com/questions/5827612/node-js-fs-readdir-recursive-directory-search
        var results = [];
        var list = fs.readdirSync(dir);
        list.forEach(function (file) {
            file = dir + '/' + file;
            var stat = fs.statSync(file);
            if (stat && stat.isDirectory()) {
                /* Recurse into a subdirectory */
                results = results.concat(Schema.getAllFiles(file));
            }
            else {
                /* Is a file */
                results.push(file);
            }
        });
        return results;
    };
    return Schema;
}());
exports.Schema = Schema;
//# sourceMappingURL=Schema.js.map