"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.propertyHasComplexTypes = exports.createPropertyDetails = exports.createPropertiesDetails = exports.createEnumSummary = exports.createPropertiesSummary = void 0;
// Copyright (c) Microsoft Corporation. All rights reserved.
// Licensed under the MIT License.
var defined_1 = require("./defined");
var defaultValue_1 = require("./defaultValue");
var SchemaClass_1 = require("../SchemaClass");
var SchemaEnum_1 = require("../SchemaEnum");
var SchemaLiteral_1 = require("../SchemaLiteral");
var mdTable = require("markdown-table");
var style = require("./style");
function createPropertiesSummary(classDefinition, knownTypes, autoLink, includeVersion, elementVersion) {
    var md = '';
    var properties = classDefinition.getAllProperties();
    if (properties !== undefined && properties.size > 0) {
        // Re-order so that required properties are up front
        // (Otherwise if there's an inherited required property, it'll be at bottom)
        properties = sortProperties(properties);
        if (includeVersion && defined_1.defined(elementVersion) && elementVersion != "1.0") {
            md += "#### Introduced in version " + elementVersion + "\n\n";
        }
        var formattedProperties = [];
        var needsType = classDefinition.type == "AdaptiveCard";
        if (!needsType) {
            classDefinition.getAllExtended().forEach(function (extended) {
                if (extended.type == "Element"
                    || extended.type == "Action"
                    || extended.type == "Inline") {
                    needsType = true;
                }
            });
        }
        if (needsType) {
            var formattedTypeProperty = {
                Property: "**type**",
                Type: "`\"" + classDefinition.type + "\"`",
                Required: "Yes",
                Description: "Must be `\"" + classDefinition.type + "\"`."
            };
            if (includeVersion) {
                formattedTypeProperty.Version = elementVersion;
            }
            formattedProperties.push(formattedTypeProperty);
        }
        var firstInheritedFormattedProperty;
        properties.forEach(function (property, name) {
            var summary = getPropertySummary(property, knownTypes, autoLink, elementVersion);
            var formattedProperty = {
                Property: style.propertyNameSummary(name),
                Type: summary.formattedType,
                Required: summary.required,
                Description: summary.description
            };
            // If we haven't reached the start of the inherited properties...
            // Note that inherited required properties are moved up to the front
            if (!firstInheritedFormattedProperty && !property.required && !classDefinition.properties.has(property.name)) {
                firstInheritedFormattedProperty = formattedProperty;
            }
            // Special case version property on Adaptive Card to be required.
            // It's actually not required in ShowCard, so in schema it's not required,
            // but for docs we want to show it as required
            if (classDefinition.type == "AdaptiveCard" && name == "version") {
                formattedProperty.Required = "Yes";
            }
            if (includeVersion) {
                formattedProperty.Version = summary.version;
            }
            formattedProperties.push(formattedProperty);
        });
        var mainFormattedProperties = [];
        var inheritedFormattedProperties = [];
        var reachedInherited = false;
        formattedProperties.forEach(function (formattedProperty) {
            if (reachedInherited || formattedProperty == firstInheritedFormattedProperty) {
                reachedInherited = true;
                inheritedFormattedProperties.push(formattedProperty);
            }
            else {
                mainFormattedProperties.push(formattedProperty);
            }
        });
        // Format as markdown table
        if (mainFormattedProperties.length > 0) {
            md += createTable(mainFormattedProperties);
            md += '\n';
        }
        if (inheritedFormattedProperties.length > 0) {
            md += "\n**Inherited properties**\n\n";
            md += createTable(inheritedFormattedProperties);
            md += "\n";
        }
    }
    return md;
}
exports.createPropertiesSummary = createPropertiesSummary;
function createTable(formattedProperties) {
    // Data needs to be formatted as follows for use with markdown library
    /*
        table([
        ['Branch', 'Commit'],
        ['main', '0123456789abcdef'],
        ['staging', 'fedcba9876543210']
        ])
    */
    var tableData = [];
    var headerRow = [];
    for (var propName in formattedProperties[0]) {
        headerRow.push(propName);
    }
    tableData.push(headerRow);
    formattedProperties.forEach(function (formattedProperty) {
        var dataRow = [];
        for (var propName in formattedProperty) {
            dataRow.push(formattedProperty[propName]);
        }
        tableData.push(dataRow);
    });
    // Format as markdown table
    return mdTable(tableData, {
        pad: false
    });
}
function sortProperties(properties) {
    var sorted = new Map();
    // Place required properties first
    properties.forEach(function (prop, name) {
        if (prop.required) {
            sorted.set(name, prop);
        }
    });
    // Then place non-required properties
    properties.forEach(function (prop, name) {
        if (!prop.required) {
            sorted.set(name, prop);
        }
    });
    return sorted;
}
function createEnumSummary(enumType) {
    var md = '';
    var rows = [];
    enumType.values.forEach(function (val) {
        var row = {
            Value: "`\"" + val.value + "\"`",
            Description: val.description
        };
        rows.push(row);
    });
    // Data needs to be formatted as follows for use with markdown library
    /*
        table([
        ['Branch', 'Commit'],
        ['main', '0123456789abcdef'],
        ['staging', 'fedcba9876543210']
        ])
    */
    var tableData = [];
    var headerRow = [];
    for (var propName in rows[0]) {
        headerRow.push(propName);
    }
    tableData.push(headerRow);
    rows.forEach(function (row) {
        var dataRow = [];
        for (var propName in row) {
            dataRow.push(row[propName]);
        }
        tableData.push(dataRow);
    });
    // Format as markdown table
    md += mdTable(tableData, {
        pad: false
    });
    md += '\n';
    return md;
}
exports.createEnumSummary = createEnumSummary;
function getPropertySummary(property, knownTypes, autoLink, elementVersion) {
    var types = property.original.type.split("|");
    property.shorthands.forEach(function (shorthand) {
        types.push.apply(types, shorthand.original.type.split("|"));
    });
    // Format the types
    var formattedTypes = [];
    types.forEach(function (type) {
        formattedTypes.push(style.type(type));
    });
    var formattedType = formattedTypes.join(", ");
    var description = property.description;
    var required;
    if (defined_1.defined(property.required) && (property.required)) {
        required = 'Yes';
    }
    else {
        var propertyDefault = property.default;
        if (defined_1.defined(propertyDefault)) {
            required = 'No, default: ' + style.defaultValue(propertyDefault);
        }
        else {
            required = 'No';
        }
    }
    var version = defaultValue_1.defaultValue(defaultValue_1.defaultValue(property.original.version, elementVersion), "1.0");
    var topLevelVersion = version;
    var includeShorthandVersions = false;
    property.shorthands.forEach(function (shorthand) {
        var shorthandVersion = defaultValue_1.defaultValue(shorthand.original.version, topLevelVersion);
        if (includeShorthandVersions || topLevelVersion != shorthandVersion) {
            version += ", " + shorthandVersion;
            includeShorthandVersions = true;
        }
    });
    return {
        type: types.join("|"),
        formattedType: formattedType,
        description: description,
        required: required,
        version: version
    };
}
function replacePipes(type) {
    return type.replace("|", ", ");
}
function createPropertiesDetails(classDefinition, headerLevel, knownTypes, autoLink, includeVersion, elementVersion) {
    var md = '';
    var properties = classDefinition.getAllProperties();
    properties.forEach(function (property, name) {
        md += createPropertyDetails(property, headerLevel, knownTypes, autoLink, includeVersion, elementVersion);
    });
    return md + '\n';
}
exports.createPropertiesDetails = createPropertiesDetails;
function createPropertyDetails(property, headerLevel, knownTypes, autoLink, includeVersion, elementVersion) {
    var md = '';
    var summary = getPropertySummary(property, knownTypes, autoLink, elementVersion);
    var type = summary.type;
    md += style.getHeaderMarkdown(property.name, headerLevel); // Includes ending newlines
    // TODO: Add plugin point for custom JSON schema properties like gltf_*
    var detailedDescription = property.description;
    if (defined_1.defined(detailedDescription)) {
        md += detailedDescription + '\n\n';
    }
    else if (defined_1.defined(summary.description)) {
        md += summary.description + '\n\n';
    }
    md += '* ' + style.propertyDetails('Type') + ': ' + summary.formattedType + '\n';
    if (includeVersion) {
        if (summary.version != elementVersion) {
            md += "* " + style.propertyDetails("Version") + " : " + summary.version + "\n";
        }
    }
    md += '* ' + style.propertyDetails('Required') + ': ' + summary.required + '\n';
    var hasComplexTypes = propertyHasComplexTypes(property);
    if (hasComplexTypes) {
        var allowedValues = getAllowedValuesForProperty(property, summary, includeVersion);
        if (allowedValues.length > 0) {
            md += '* ' + style.propertyDetails('Allowed values') + ':';
            allowedValues.forEach(function (allowedValue) {
                md += "\n  * " + allowedValue;
            });
        }
    }
    return md + '\n';
}
exports.createPropertyDetails = createPropertyDetails;
function propertyHasComplexTypes(property) {
    var answer = false;
    var allTypes = property.getAllTypes();
    allTypes.forEach(function (propertyType) {
        if (propertyType.type instanceof SchemaClass_1.SchemaClass || propertyType.type instanceof SchemaEnum_1.SchemaEnum) {
            answer = true;
        }
    });
    return answer;
}
exports.propertyHasComplexTypes = propertyHasComplexTypes;
function getAllowedValuesForProperty(property, summary, includeVersion) {
    var allowedValues = [];
    property.getAllTypes().forEach(function (propertyType) {
        allowedValues.push.apply(allowedValues, getAllowedValuesForPropertyType(propertyType, summary, includeVersion));
    });
    return allowedValues;
}
function getAllowedValuesForPropertyType(propertyType, summary, includeVersion) {
    var allowedValues = [];
    if (propertyType.type instanceof SchemaEnum_1.SchemaEnum) {
        propertyType.type.values.forEach(function (enumValue) {
            var descriptions = [];
            if (includeVersion) {
                if (enumValue.original.version && enumValue.original.version !== summary.version) {
                    descriptions.push("Added in version " + enumValue.original.version + ".");
                }
            }
            if (enumValue.description && enumValue.description.length > 0) {
                descriptions.push(enumValue.description);
            }
            if (descriptions.length > 0) {
                allowedValues.push(style.enumValue(enumValue.value) + ": " + descriptions.join(" "));
            }
            else {
                allowedValues.push(style.enumValue(enumValue.value));
            }
        });
    }
    else if (propertyType.type instanceof SchemaClass_1.SchemaClass) {
        allowedValues.push.apply(allowedValues, getAllowedValuesForClass(propertyType.type, summary, includeVersion));
    }
    else if (propertyType.type instanceof SchemaLiteral_1.SchemaLiteral) {
        allowedValues.push(style.type(propertyType.type.type));
    }
    else {
        throw new Error("Unknown type");
    }
    return allowedValues;
}
function getAllowedValuesForClass(classType, summary, includeVersion) {
    var answer = [];
    if (classType.implementations.length > 0) {
        classType.implementations.forEach(function (implementation) {
            answer.push.apply(answer, getAllowedValuesForClass(implementation, summary, includeVersion));
        });
    }
    else {
        answer.push(style.type(classType.type));
        if (classType.shorthand) {
            answer.push.apply(answer, getAllowedValuesForProperty(classType.shorthand, summary, includeVersion));
        }
    }
    return answer;
}
//# sourceMappingURL=generate-markdown.js.map