const vueDocs = require('vue-docgen-api');
const path = require('path');
const _ = require('lodash');
const fs = require('fs');
var async = require('async');
var genShared = require('./gen-shared');

//特殊的标记tag，不需要转成属性
const ignoreTags = ['ignore', 'jsonValues']
function tag(obj, tagName) {
  return obj.tags && obj.tags[tagName] && obj.tags[tagName][0].description;
}
function tag(obj, tagName) {
  return obj.tags && obj.tags[tagName] && obj.tags[tagName][0].description;
}
/** 事件的tag数据格式和其他的不一样 */
function eventTag(obj, tagName) {
  let tagDef = _.find(obj.tags, t=> t.title === tagName);
  return tagDef && tagDef.content;
}
/** 将doc的属性的tags转成对象模式 */
function tagsToMap(obj){
  let props = {}
  _.forIn(obj.tags,(v,tagName)=>{
    if(!_.includes(ignoreTags, tagName)){
      let tagValue = v[0].description
      props[tagName] = tagValue
    }
  })
  return props
}
function run(components, {srcFolder,groupFolder}, inspect) {
  let schemas = {};
  let enums = {};
  console.log('正在生成组件schema文件...');
  async.forEachOf(
    components,
    (filePath, name, callback) => {
      vueDocs.parse(filePath).then((doc) => {
        if (inspect) {
          console.log(`${name}的原始定义: `);
          console.log(JSON.stringify(doc, null, 2));
        }
        const comAttrs = tagsToMap(doc);
        let schema = {
          description: doc.description,
          defType: 'component',
          ...comAttrs
        };
        let properties = {};
        _.each(doc.props, (p) => {
          if (tag(p, 'ignore')) {
            return;
          }
          let _p = {
            type: null,
            description: null
          };
          let defaultValue = p.defaultValue && p.defaultValue.value;
          if (_.startsWith(defaultValue, "'")) {
            _p.default = _.trim(defaultValue, "'");
          } else if (defaultValue === 'false' || defaultValue === 'true') {
            _p.default = defaultValue === 'false' ? false : true;
          } else if (new RegExp(/\d+/).test(defaultValue)) {
            _p.default = _.toNumber(defaultValue);
          }
          if (p.description) {
            _p.description = p.description;
          }
          let type = p.type && p.type.name;
          if (type) {
            //复合类型取第一个类型
            let verticalLineIndex = type.indexOf('|');
            if (verticalLineIndex > 0) {
              type = type.substring(0, verticalLineIndex);
            }
            if (type === 'array') {
              _p.type = 'object[]';
            } else {
              _p.type = type;
            }
          } else {
            _p.type = 'string';
          }
          //选项类型读取values构造枚举类型
          let jsonValues = tag(p, 'jsonValues');
          if (p.values || jsonValues) {
            let enumType = `${name}${_.upperFirst(p.name)}Values`;
            //@values red, blue, green
            let _values = null;
            if (p.values) {
              _values = p.values;
            } else if (jsonValues) {
              //@jsonValues [{"value": "small","title":"小"}, {"value": "medium","title":"中"}, {"value": "large","title":"大"}]
              //json数组格式的values需要转换为真实的数组格式
              _values = JSON.parse(jsonValues);
            }
            enums[enumType] = _values;
            _p.type = enumType;
          }
          const pAttrs = tagsToMap(p);
          _p = {..._p, ...pAttrs};
          properties[p.name] = _p;
        });
        schema.properties = properties;
        let slots = {};
        _.each(doc.slots, (_slot) => {
          let ignore = tag(_slot, 'ignore');
          if (ignore || !_slot.description) {
            return;
          }
          let _p = {
            title: _slot.description,
            context: {
              name: 'slotProps',
              properties: {}
            },
            turnOn: tag(_slot, 'turnOn')
          };
          slots[_slot.name] = _p;
        });
        schema.slots = slots;
        
        let events = {};
        _.each(doc.events, (evt) => {
          let ignore = eventTag(evt, 'ignore');
          if (ignore || !evt.description) {
            return;
          }
          events[evt.name] = {
            title: evt.description,
            properties: {}
          };
        });
        schema.events = events;
        schemas[name] = schema;

        callback();
      });
    },
    (err) => {
      if (err) {
        return;
      }
      if (_.isEmpty(schemas)) {
        console.log('没有任何组件schema需要生成');
        return;
      }

      //生成共享schema方便各组件使用
      let sharedFolder = path.join(srcFolder, 'shared');
      genShared.run(sharedFolder);

      let elementsFolder = path.join(srcFolder, groupFolder);
      //如果schema的源码目录elements目录不存在，先创建出来
      if (!fs.existsSync(elementsFolder)) {
        fs.mkdirSync(elementsFolder, { recursive: true });
      }
      _.forIn(schemas, (schema, name) => {
        schema.properties = schema.properties || {};
        let elementFile = path.join(elementsFolder, `${name}.json`);
        //读取已经存在的schema文件做合并
        if (fs.existsSync(elementFile)) {
          let oldElementSchema = fs.readFileSync(elementFile);
          oldElementSchema = JSON.parse(oldElementSchema);
          oldElementSchema.properties = oldElementSchema.properties || {};
          _.merge(oldElementSchema, _.cloneDeep(schema));
          schema = oldElementSchema;
        }
        fs.writeFileSync(elementFile, JSON.stringify(schema, null, 2));
      });
      if (_.isEmpty(enums)) {
        console.log('组件schema文件生成完毕');
        return;
      }
      let enumsFolder = path.join(srcFolder, 'enums');
      if (!fs.existsSync(enumsFolder)) {
        fs.mkdirSync(enumsFolder, { recursive: true });
      }
      _.forIn(enums, (values, name) => {
        let enumFile = path.join(enumsFolder, `${name}.json`);
        let enumContent = {
          $schema: '',
          classType: 'Enum'
        };
        //读取已经存在的enum文件做合并
        if (fs.existsSync(enumFile)) {
          let oldEnum = fs.readFileSync(enumFile);
          oldEnum = JSON.parse(oldEnum);
          oldEnum.values = values;
          enumContent = oldEnum;
        } else {
          enumContent.values = values;
        }
        fs.writeFileSync(enumFile, JSON.stringify(enumContent, null, 2));
      });
      console.log('组件schema文件生成完毕');
    }
  );
}

module.exports = {
  run
};
